import binascii
import ipaddress
import uuid
from calendar import timegm
from datetime import datetime, timezone

import cbor2
from marshmallow import fields as m_fields, utils


# Fields for custom tags (not handled natively by cbor2)


class Tagged(m_fields.Field):
    """Wrap a field in a CBOR Tag

    :param tagged_field: Any field instance including ``fields.Nested`` schemas
    :param tag: Tag ID.
    """

    default_error_messages = {'wrong_tag': 'unexpected tag'}

    def __init__(self, tagged_field, *, tag, **kwargs):
        if not isinstance(tag, int) or tag < 0:
            raise ValueError('tag must be an int > 0')
        self._tagged_field = tagged_field
        self._tag = tag
        super().__init__(**kwargs)

    def _serialize(self, nested_obj, attr, obj, **kwargs):
        serialized = self._tagged_field._serialize(nested_obj, attr, obj, **kwargs)
        return cbor2.CBORTag(self._tag, serialized)

    def _deserialize(self, value, attr, data, partial=None, **kwargs):
        if isinstance(value, cbor2.CBORTag):
            if value.tag == self._tag:
                value = value.value
            else:
                self.make_error('wrong_tag', input=value)
        return self._tagged_field._deserialize(
            value, attr, data, partial=partial, **kwargs
        )


class Embedded(m_fields.Field):
    """Serialize a field as CBOR bytes

    :param embedded_field: Any field instance including ``fields.Nested`` schemas
    """

    def __init__(self, embedded_field, **kwargs):
        self._embedded_field = embedded_field
        super().__init__(**kwargs)

    def _serialize(self, nested_obj, attr, obj, **kwargs):
        serialized = self._embedded_field._serialize(nested_obj, attr, obj, **kwargs)
        return cbor2.dumps(serialized)

    def _deserialize(self, value, attr, data, partial=None, **kwargs):
        if isinstance(value, bytes):
            value = cbor2.loads(value)
        return self._embedded_field._deserialize(
            value, attr, data, partial=partial, **kwargs
        )


# Native CBOR fields that can just be passed through


class Bytes(m_fields.Field):
    """Instead of using Python's raw bytes type to represent CBOR bytes
    we can chose to load them as strings. If you don't want to interpret
    bytes objects just use ``fields.Raw()`` instead.

    :param load_as: String representation of bytes object
    """

    LOAD_AS = {
        None: lambda x: x,
        'hex': lambda x: binascii.hexlify(x).decode(),
        'utf8': lambda x: x.decode('utf8', errors='backslashescape'),
    }
    DUMP_AS = {
        None: lambda x: x,
        'hex': lambda x: binascii.unhexlify(x),
        'utf8': lambda x: x.encode('utf8'),
    }

    def __init__(self, *, load_as=None, **kwargs):
        try:
            self._load_func = self.LOAD_AS[load_as]
            self._dump_func = self.DUMP_AS[load_as]
        except KeyError:
            raise ValueError(f'unsupported bytes representation {load_as}')
        super().__init__(**kwargs)

    def _serialize(self, value, attr, obj, **kwargs):
        return self._dump_func(value)

    def _deserialize(self, value, attr, data, **kwargs):
        return self._load_func(value)


class Timestamp(m_fields.AwareDateTime):
    """Unix epoch seconds in UTC.

    If you want it to be on the wire as an CBOR Tagged timestamp use
    fields.Tagged(Timestamp(), tag=1)
    """

    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, datetime):
            if not value.microsecond:
                return timegm(value.utctimetuple())
            else:
                return timegm(value.utctimetuple()) + value.microsecond / 1000000
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, datetime):
            return value
        elif isinstance(value, str):
            return super()._deserialize(value, attr, data, **kwargs)
        else:
            return datetime.fromtimestamp(value, tz=timezone.utc)


class AwareDateTime(m_fields.AwareDateTime):
    _serialize = m_fields.Field._serialize

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, datetime):
            return value
        else:
            return super()._deserialize(value, attr, data, **kwargs)


class UUID(m_fields.UUID):
    _serialize = m_fields.Field._serialize

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, uuid.UUID):
            return value
        else:
            return super()._deserialize(value, attr, data, **kwargs)


class IP(m_fields.IP):
    _serialize = m_fields.Field._serialize

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, self.DESERIALIZATION_CLASS):
            return value
        else:
            return super()._deserialize(value, attr, data, **kwargs)


IPv4 = type('IPv4', (m_fields.IPv4, IP), {})
IPv6 = type('IPv6', (m_fields.IPv6, IP), {})
IPv4Interface = m_fields.IPv4Interface
IPv6Interface = m_fields.IPv6Interface


class IPNetwork(m_fields.Field):
    default_error_messages = {"invalid_ip_network": "Not a valid IP Network."}

    DESERIALIZATION_CLASS = None

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None
        if isinstance(value, self.DESERIALIZATION_CLASS):
            return value
        try:
            return (self.DESERIALIZATION_CLASS or ipaddress.ip_interface)(
                utils.ensure_text_type(value)
            )
        except (ValueError, TypeError) as error:
            raise self.make_error("invalid_ip_network") from error


class IPv4Network(IPNetwork):
    """A IPv4 Network field. (without the host part)"""

    default_error_messages = {"invalid_ip_network": "Not a valid IPv4 network."}

    DESERIALIZATION_CLASS = ipaddress.IPv4Network


class IPv6Network(IPNetwork):
    """A IPv6 Network field. (without the host part)"""

    default_error_messages = {"invalid_ip_network": "Not a valid IPv6 network."}

    DESERIALIZATION_CLASS = ipaddress.IPv6Network


# Unchanged fields from marshmallow

Field = m_fields.Field
Raw = m_fields.Raw
Nested = m_fields.Nested
Mapping = m_fields.Mapping
Dict = m_fields.Dict
List = m_fields.List
Tuple = m_fields.Tuple
String = m_fields.String
Number = m_fields.Number
Integer = m_fields.Integer
Decimal = m_fields.Decimal
Boolean = m_fields.Boolean
Float = m_fields.Float
Time = m_fields.Time
Date = m_fields.Date
TimeDelta = m_fields.TimeDelta
Url = m_fields.Url
URL = m_fields.URL
Email = m_fields.Email
Method = m_fields.Method
Function = m_fields.Function
Str = m_fields.Str
Bool = m_fields.Bool
Int = m_fields.Int
Constant = m_fields.Constant
Pluck = m_fields.Pluck



from marshmallow_cbor import Schema, fields


class PersonSchema(Schema):
    """
    Simple data with string keys and values, Can be represented with JSON. No difference
    from regular Marshmallow

    >>> from binascii import hexlify
    >>> person = {
    ...     "name": "Simon",
    ...     "occupation": "Skydiving Instructor",
    ...     "homepage": "https://example.com/letsgoskydiving"
    ... }
    >>> expected = (b'a3646e616d656553696d6f6e6a6f636375706174696f6e74536b79646976696e672'
    ... b'0496e7374727563746f7268686f6d6570616765782368747470733a2f2f6578616d706c652e636f'
    ... b'6d2f6c657473676f736b79646976696e67')
    >>> schema = PersonSchema()
    >>> hexlify(schema.dumps(person)) == expected
    True
    """

    name = fields.String()
    occupation = fields.String()
    homepage = fields.Url()



import datetime as dt
import ipaddress
from binascii import hexlify, unhexlify

import cbor2
import pytest

from marshmallow_cbor import Schema, fields


def dumper(data, **kwargs):
    encoded = cbor2.dumps(data, **kwargs)
    return hexlify(encoded).decode()


# DateTime


class DTSchema(Schema):
    ts = fields.AwareDateTime()


class TSSchema(Schema):
    ts = fields.Tagged(fields.Timestamp(), tag=1)


# IP addresses


class IPSchema(Schema):
    ip = fields.IPv4()


class NetSchema(Schema):
    ip = fields.IPv4Network()


# Bytes as strings


class HexSchema(Schema):
    data = fields.Bytes(load_as="hex")


class Utf8bytesSchema(Schema):
    data = fields.Bytes(load_as="utf8")


@pytest.mark.parametrize(
    'schema, data, expected',
    [
        (
            DTSchema(),
            {'ts': dt.datetime(2021, 6, 30, 19, tzinfo=dt.timezone.utc)},
            dumper({'ts': dt.datetime(2021, 6, 30, 19, tzinfo=dt.timezone.utc)}),
        ),
        (
            TSSchema(),
            {'ts': dt.datetime(2021, 6, 30, 19, tzinfo=dt.timezone.utc)},
            dumper(
                {'ts': dt.datetime(2021, 6, 30, 19, tzinfo=dt.timezone.utc)},
                datetime_as_timestamp=True,
            ),
        ),
        (
            IPSchema(),
            {'ip': ipaddress.IPv4Address('192.168.0.1')},
            dumper({'ip': ipaddress.IPv4Address('192.168.0.1')}),
        ),
        (
            NetSchema(),
            {'ip': ipaddress.IPv4Network('192.168.0.0/24')},
            dumper({'ip': ipaddress.IPv4Network('192.168.0.0/24')}),
        ),
        (HexSchema(), {'data': '010203'}, dumper({'data': b'\x01\x02\x03'})),
        (
            Utf8bytesSchema(),
            {'data': '\u0001\u0002\u0003'},
            dumper({'data': b'\x01\x02\x03'}),
        ),
    ],
)
def test_field(schema, data, expected):
    assert hexlify(schema.dumps(data)).decode() == expected
    assert schema.loads(unhexlify(expected)) == data



from marshmallow_cbor import Schema, fields


class PointSchema(Schema):
    """
    Simple tagged schema, this as set as an option on the Meta class

    >>> from binascii import hexlify, unhexlify
    >>> point = {'x': 1.2, 'y': 1.3, 'z': 1.4}
    >>> schema = PointSchema()
    >>> hexlify(schema.dumps(point))
    b'd93039a36178fb3ff33333333333336179fb3ff4cccccccccccd617afb3ff6666666666666'
    >>> schema.loads(unhexlify(_))
    {'x': 1.2, 'y': 1.3, 'z': 1.4}
    """

    x = fields.Float()
    y = fields.Float()
    z = fields.Float()

    class Meta:
        tag = 12345



"""
Experimenting with RFC 8392 CBOR Web Tokens

>>> import binascii
>>> from pprint import pprint
>>> schema = CWTClaimsSchema()
>>> claims = schema.loads(
...     binascii.unhexlify(
...         b'a70175636f61703a2f2f61732e6578616d706c652e636f6d02656572696b7703'
...         b'7818636f61703a2f2f6c696768742e6578616d706c652e636f6d041a5612aeb0'
...         b'051a5610d9f0061a5610d9f007420b71'
...     )
... )
>>> pprint(claims)
{'aud': 'coap://light.example.com',
 'cti': '0b71',
 'exp': datetime.datetime(2015, 10, 5, 17, 9, 4, tzinfo=datetime.timezone.utc),
 'iat': datetime.datetime(2015, 10, 4, 7, 49, 4, tzinfo=datetime.timezone.utc),
 'iss': 'coap://as.example.com',
 'nbf': datetime.datetime(2015, 10, 4, 7, 49, 4, tzinfo=datetime.timezone.utc),
 'sub': 'erikw'}
>>> macd_cwt = binascii.unhexlify(
...     b'd83dd18443a10104a1044c53796d6d65747269633235365850a70175636f6170'
...     b'3a2f2f61732e6578616d706c652e636f6d02656572696b77037818636f61703a'
...     b'2f2f6c696768742e6578616d706c652e636f6d041a5612aeb0051a5610d9f006'
...     b'1a5610d9f007420b7148093101ef6d789200'
... )
>>> macd_schema = CWTMACSchema()
>>> macd_claims = macd_schema.loads(macd_cwt)
>>> pprint(macd_claims)
{'alg': {'alg': 4},
 'kid': {'kid': 'Symmetric256'},
 'payload': {'aud': 'coap://light.example.com',
         'cti': '0b71',
         'exp': datetime.datetime(2015, 10, 5, 17, 9, 4, tzinfo=datetime.timezone.utc),
         'iat': datetime.datetime(2015, 10, 4, 7, 49, 4, tzinfo=datetime.timezone.utc),
         'iss': 'coap://as.example.com',
         'nbf': datetime.datetime(2015, 10, 4, 7, 49, 4, tzinfo=datetime.timezone.utc),
         'sub': 'erikw'},
 'tag': '093101ef6d789200'}
"""

import cbor2
from marshmallow import pre_load, post_dump, ValidationError
from marshmallow_cbor import Schema, fields


class CWTClaimsSchema(Schema):
    iss = fields.Url(data_key=1, schemes=('coap',))
    sub = fields.String(data_key=2)
    aud = fields.Url(data_key=3, schemes=('coap',))
    exp = fields.Timestamp(data_key=4)
    nbf = fields.Timestamp(data_key=5)
    iat = fields.Timestamp(data_key=6)
    cti = fields.Bytes(data_key=7, load_as='hex')


class AlgSchema(Schema):
    alg = fields.Integer(data_key=1)


class KidSchema(Schema):
    kid = fields.Bytes(data_key=4, load_as='utf8')


class CWTMACSchema(Schema):
    alg = fields.Embedded(fields.Nested(AlgSchema))            # Protected
    kid = fields.Nested(KidSchema)                             # Unprotected
    payload = fields.Embedded(fields.Nested(CWTClaimsSchema))  # Protected
    tag = fields.Bytes(load_as='hex')                          # Unprotected

    @pre_load
    def unpack_tags(self, data, **kwargs):
        if isinstance(data, cbor2.CBORTag) and data.tag == 61 and data.value.tag == 17:
            return data.value.value
        else:
            raise ValidationError

    def _deserialize(self, data, many, **kwargs):
        row = dict(alg=data[0], kid=data[1], payload=data[2], tag=data[3])
        return super()._deserialize(row, many=many, **kwargs)

    def _serialize(self, value, many, **kwargs):
        value = super()._serialize(value, many=many, **kwargs)
        payload = [value['alg'], value['kid'], value['payload'], value['tag']]
        return payload

    @post_dump
    def pack_tags(self, data, **kwargs):
        return cbor2.CBORTag(61, cbor2.CBORTag(17, data))



import cbor2
from marshmallow import Schema as mSchema, SchemaOpts
from marshmallow.validate import ValidationError


class CBOROptions(SchemaOpts):
    def __init__(self, meta, **kwargs):
        SchemaOpts.__init__(self, meta, **kwargs)
        self.tag = getattr(meta, "tag", None)
        self.render_module = cbor2


class Schema(mSchema):
    OPTIONS_CLASS = CBOROptions

    def _deserialize(self, data, many, **kwargs):
        if isinstance(data, cbor2.CBORTag):
            if data.tag == self.opts.tag:
                data = data.value
            else:
                raise ValidationError(f'unexpected tag: {data.tag}')
        return super()._deserialize(data, many=many, **kwargs)

    def _serialize(self, value, many, **kwargs):
        value = super()._serialize(value, many=many, **kwargs)
        if self.opts.tag:
            value = cbor2.CBORTag(self.opts.tag, value)
        return value



from setuptools import setup


kwargs = {}

setup(
    use_scm_version={'version_scheme': 'post-release', 'local_scheme': 'dirty-tag'},
    setup_requires=['setuptools >= 40.7.0', 'setuptools_scm >= 1.7.0'],
    **kwargs
)



import binascii
import decimal
import uuid

import pytest

from cbor2 import CBORTag
from marshmallow import pre_load, post_dump, ValidationError
from marshmallow_cbor import Schema
from marshmallow_cbor.fields import (
    AwareDateTime,
    Decimal,
    UUID,
    Boolean,
    Tagged,
    Embedded,
    Nested,
)


class DateTimeSchema(Schema):
    ts = AwareDateTime(data_key=0)


class DecimalSchema(Schema):
    num = Decimal()


class UUIDSchema(Schema):
    uid = UUID()


class TaggedSchema(Schema):
    a = Boolean()
    b = Decimal()

    class Meta:
        tag = 4096


class EmbedSchema(Schema):
    a = Boolean()
    b = Decimal()


class WithEmbed(Schema):
    payload = Embedded(Nested(EmbedSchema))


class NestedTaggedSchema(Schema):
    a = Tagged(Nested(UUIDSchema), tag=3360)
    b = Boolean()


class TagWithinTagSchema(Schema):
    a = Boolean()
    b = Boolean()

    @pre_load
    def unpack_tags(self, data, **kwargs):
        if data.tag == 5990 and data.value.tag == 5991:
            return data.value.value

    @post_dump
    def pack_tags(self, data, **kwargs):
        return CBORTag(5990, CBORTag(5991, data))


@pytest.mark.parametrize(
    'schema, source, expected',
    [
        (
            DateTimeSchema(),
            {0: '2012-03-07T14:32:00.769Z'},
            b'a100c0781b323031322d30332d30375431343a33323a30302e3736393030305a',
        ),
        (
            DecimalSchema(),
            {'num': decimal.Decimal(355) / decimal.Decimal(113)},
            b'a1636e756dc482381ac24c0a26aa2d773467d3bb65f268',
        ),
        (
            UUIDSchema(),
            {'uid': uuid.uuid5(uuid.NAMESPACE_DNS, 'example.com')},
            b'a163756964d82550cfbff0d193755685968c48ce8b15ae17',
        ),
        (
            TaggedSchema(),
            {'a': True, 'b': decimal.Decimal(1)},
            b'd91000a26162c48200016161f5',
        ),
        (
            WithEmbed(),
            {'payload': {'a': True, 'b': decimal.Decimal(1)}},
            b'a1677061796c6f61644aa26162c48200016161f5',
        ),
        (
            NestedTaggedSchema(),
            {'a': {'uid': uuid.uuid5(uuid.NAMESPACE_DNS, 'example.com')}, 'b': True},
            b'a26162f56161d90d20a163756964d82550cfbff0d193755685968c48ce8b15ae17',
        ),
        (
            TagWithinTagSchema(),
            CBORTag(5990, CBORTag(5991, {'a': True, 'b': False})),
            b'd91766d91767a26162f46161f5',
        ),
    ],
)
def test_schema_dumps(schema, source, expected):

    data = schema.load(source)
    print(data)

    encoded = schema.dumps(data)

    assert binascii.hexlify(encoded) == expected

    data2 = schema.loads(encoded)

    assert data == data2


class IncorrectTagSchema(Schema):
    a = Boolean()
    b = Boolean()

    class Meta:
        tag = 4096


class IncorrectNestedTag(Schema):
    a = Tagged(Nested(UUIDSchema), tag=3017)
    b = Boolean()


@pytest.mark.parametrize(
    'schema, source, expected',
    [
        (IncorrectTagSchema(), b'D91001A26161F56162F4', ValidationError),
        (
            IncorrectNestedTag(),
            b'a26162f56161d90d20a163756964d82550cfbff0d193755685968c48ce8b15ae17',
            ValidationError,
        ),
    ],
)
def test_exceptions(schema, source, expected):
    with pytest.raises(expected):
        schema.loads(binascii.unhexlify(source))
